import argparse
import codecs
import os
from datetime import datetime, timezone
from operator import itemgetter
from typing import Any, List

from colorama import Fore, Style

from frida_tools.application import ConsoleApplication

STYLE_DIR = Fore.BLUE + Style.BRIGHT
STYLE_EXECUTABLE = Fore.GREEN + Style.BRIGHT
STYLE_LINK = Fore.CYAN + Style.BRIGHT
STYLE_ERROR = Fore.RED + Style.BRIGHT


def main() -> None:
    app = LsApplication()
    app.run()


class LsApplication(ConsoleApplication):
    def _add_options(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("files", help="files to list information about", nargs="*")

    def _usage(self) -> str:
        return "%(prog)s [options] [FILE]..."

    def _initialize(self, parser: argparse.ArgumentParser, options: argparse.Namespace, args: List[str]) -> None:
        self._files = options.files

    def _needs_target(self) -> bool:
        return False

    def _start(self) -> None:
        try:
            self._attach(self._pick_worker_pid())

            data_dir = os.path.dirname(__file__)
            with codecs.open(os.path.join(data_dir, "fs_agent.js"), "r", "utf-8") as f:
                source = f.read()

            def on_message(message: Any, data: Any) -> None:
                print(message)

            assert self._session is not None
            script = self._session.create_script(name="ls", source=source)
            script.on("message", on_message)
            self._on_script_created(script)
            script.load()

            groups = script.exports_sync.ls(self._files)
        except Exception as e:
            self._update_status(f"Failed to retrieve listing: {e}")
            self._exit(1)
            return

        exit_status = 0
        for i, group in enumerate(sorted(groups, key=lambda g: g["path"])):
            path = group["path"]
            if path != "" and len(groups) > 1:
                if i > 0:
                    self._print("")
                self._print(path + ":")

            for path, message in group["errors"]:
                self._print(STYLE_ERROR + message + Style.RESET_ALL)
                exit_status = 2

            rows = []
            for name, target, type, access, nlink, owner, group, size, raw_mtime in group["entries"]:
                mtime = datetime.fromtimestamp(raw_mtime / 1000.0, tz=timezone.utc)
                rows.append((type + access, str(nlink), owner, group, str(size), mtime.strftime("%c"), name, target))

            if len(rows) == 0:
                break

            widths = []
            for column_index in range(len(rows[0]) - 2):
                width = max(map(lambda row: len(row[column_index]), rows))
                widths.append(width)

            adjustments = [
                "",
                ">",
                "<",
                "<",
                ">",
                "<",
            ]
            col_formats = []
            for i, width in enumerate(widths):
                adj = adjustments[i]
                if adj != "":
                    fmt = "{:" + adj + str(width) + "}"
                else:
                    fmt = "{}"
                col_formats.append(fmt)
            row_description = " ".join(col_formats)

            for row in sorted(rows, key=itemgetter(6)):
                meta_fields = row_description.format(*row[:-2])

                name, target = row[6:8]
                ftype_and_perms = row[0]
                ftype = ftype_and_perms[0]
                fperms = ftype_and_perms[1:]
                name = format_name(name, ftype, fperms, target)

                self._print(meta_fields + " " + name)

        self._exit(exit_status)


def format_name(name: str, ftype: str, fperms: str, target) -> str:
    if ftype == "l":
        target_path, target_details = target
        if target_details is not None:
            target_type, target_perms = target_details
            target_summary = format_name(target_path, target_type, target_perms, None)
        else:
            target_summary = STYLE_ERROR + target_path + Style.RESET_ALL
        return STYLE_LINK + name + Style.RESET_ALL + " -> " + target_summary

    if ftype == "d":
        return STYLE_DIR + name + Style.RESET_ALL

    if "x" in fperms:
        return STYLE_EXECUTABLE + name + Style.RESET_ALL

    return name


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



def main() -> None:
    import argparse
    import functools
    import json
    import math
    import platform
    import sys
    from base64 import b64encode
    from typing import List, Tuple, Union

    try:
        import termios
        import tty
    except:
        pass

    import frida._frida as _frida

    from frida_tools.application import ConsoleApplication

    class PSApplication(ConsoleApplication):
        def _add_options(self, parser: argparse.ArgumentParser) -> None:
            parser.add_argument(
                "-a",
                "--applications",
                help="list only applications",
                action="store_true",
                dest="list_only_applications",
                default=False,
            )
            parser.add_argument(
                "-i",
                "--installed",
                help="include all installed applications",
                action="store_true",
                dest="include_all_applications",
                default=False,
            )
            parser.add_argument(
                "-j",
                "--json",
                help="output results as JSON",
                action="store_const",
                dest="output_format",
                const="json",
                default="text",
            )

        def _initialize(self, parser: argparse.ArgumentParser, options: argparse.Namespace, args: List[str]) -> None:
            if options.include_all_applications and not options.list_only_applications:
                parser.error("-i cannot be used without -a")
            self._list_only_applications = options.list_only_applications
            self._include_all_applications = options.include_all_applications
            self._output_format = options.output_format
            self._terminal_type, self._icon_size = self._detect_terminal()

        def _usage(self) -> str:
            return "%(prog)s [options]"

        def _start(self) -> None:
            if self._list_only_applications:
                self._list_applications()
            else:
                self._list_processes()

        def _list_processes(self) -> None:
            if self._output_format == "text" and self._terminal_type == "iterm2":
                scope = "full"
            else:
                scope = "minimal"

            try:
                assert self._device is not None
                processes = self._device.enumerate_processes(scope=scope)
            except Exception as e:
                self._update_status(f"Failed to enumerate processes: {e}")
                self._exit(1)
                return

            if self._output_format == "text":
                if len(processes) > 0:
                    pid_column_width = max(map(lambda p: len(str(p.pid)), processes))
                    icon_width = max(map(compute_icon_width, processes))
                    name_column_width = icon_width + max(map(lambda p: len(p.name), processes))

                    header_format = "%" + str(pid_column_width) + "s  %s"
                    self._print(header_format % ("PID", "Name"))
                    self._print(f"{pid_column_width * '-'}  {name_column_width * '-'}")

                    line_format = "%" + str(pid_column_width) + "d  %s"
                    name_format = "%-" + str(name_column_width - icon_width) + "s"

                    for process in sorted(processes, key=functools.cmp_to_key(compare_processes)):
                        if icon_width != 0:
                            icons = process.parameters.get("icons", None)
                            if icons is not None:
                                icon = self._render_icon(icons[0])
                            else:
                                icon = "   "
                            name = icon + " " + name_format % process.name
                        else:
                            name = name_format % process.name

                        self._print(line_format % (process.pid, name))
                else:
                    self._log("error", "No running processes.")
            elif self._output_format == "json":
                result = []
                for process in sorted(processes, key=functools.cmp_to_key(compare_processes)):
                    result.append({"pid": process.pid, "name": process.name})
                self._print(json.dumps(result, sort_keys=False, indent=2))

            self._exit(0)

        def _list_applications(self) -> None:
            if self._output_format == "text" and self._terminal_type == "iterm2":
                scope = "full"
            else:
                scope = "minimal"

            try:
                assert self._device is not None
                applications = self._device.enumerate_applications(scope=scope)
            except Exception as e:
                self._update_status(f"Failed to enumerate applications: {e}")
                self._exit(1)
                return

            if not self._include_all_applications:
                applications = list(filter(lambda app: app.pid != 0, applications))

            if self._output_format == "text":
                if len(applications) > 0:
                    pid_column_width = max(map(lambda app: len(str(app.pid)), applications))
                    icon_width = max(map(compute_icon_width, applications))
                    name_column_width = icon_width + max(map(lambda app: len(app.name), applications))
                    identifier_column_width = max(map(lambda app: len(app.identifier), applications))

                    header_format = (
                        "%"
                        + str(pid_column_width)
                        + "s  "
                        + "%-"
                        + str(name_column_width)
                        + "s  "
                        + "%-"
                        + str(identifier_column_width)
                        + "s"
                    )
                    self._print(header_format % ("PID", "Name", "Identifier"))
                    self._print(f"{pid_column_width * '-'}  {name_column_width * '-'}  {identifier_column_width * '-'}")

                    line_format = "%" + str(pid_column_width) + "s  %s  %-" + str(identifier_column_width) + "s"
                    name_format = "%-" + str(name_column_width - icon_width) + "s"

                    for app in sorted(applications, key=functools.cmp_to_key(compare_applications)):
                        if icon_width != 0:
                            icons = app.parameters.get("icons", None)
                            if icons is not None:
                                icon = self._render_icon(icons[0])
                            else:
                                icon = "   "
                            name = icon + " " + name_format % app.name
                        else:
                            name = name_format % app.name

                        if app.pid == 0:
                            self._print(line_format % ("-", name, app.identifier))
                        else:
                            self._print(line_format % (app.pid, name, app.identifier))

                elif self._include_all_applications:
                    self._log("error", "No installed applications.")
                else:
                    self._log("error", "No running applications.")
            elif self._output_format == "json":
                result = []
                if len(applications) > 0:
                    for app in sorted(applications, key=functools.cmp_to_key(compare_applications)):
                        result.append({"pid": (app.pid or None), "name": app.name, "identifier": app.identifier})
                self._print(json.dumps(result, sort_keys=False, indent=2))

            self._exit(0)

        def _render_icon(self, icon) -> str:
            return "\033]1337;File=inline=1;width={}px;height={}px;:{}\007".format(
                self._icon_size, self._icon_size, b64encode(icon["image"]).decode("ascii")
            )

        def _detect_terminal(self) -> Tuple[str, int]:
            icon_size = 0

            if not self._have_terminal or self._plain_terminal or platform.system() != "Darwin":
                return ("simple", icon_size)

            fd = sys.stdin.fileno()
            old_attributes = termios.tcgetattr(fd)
            try:
                tty.setraw(fd)
                new_attributes = termios.tcgetattr(fd)
                new_attributes[3] = new_attributes[3] & ~termios.ICANON & ~termios.ECHO
                termios.tcsetattr(fd, termios.TCSANOW, new_attributes)

                sys.stdout.write("\033[1337n")
                sys.stdout.write("\033[5n")
                sys.stdout.flush()

                response = self._read_terminal_response("n")
                if response not in ("0", "3"):
                    self._read_terminal_response("n")

                    if response.startswith("ITERM2 "):
                        version_tokens = response.split(" ", 1)[1].split(".", 2)
                        if len(version_tokens) >= 2 and int(version_tokens[0]) >= 3:
                            sys.stdout.write("\033[14t")
                            sys.stdout.flush()
                            height_in_pixels = int(self._read_terminal_response("t").split(";")[1])

                            sys.stdout.write("\033[18t")
                            sys.stdout.flush()
                            height_in_cells = int(self._read_terminal_response("t").split(";")[1])

                            icon_size = math.ceil((height_in_pixels / height_in_cells) * 1.77)

                            return ("iterm2", icon_size)

                return ("simple", icon_size)
            finally:
                termios.tcsetattr(fd, termios.TCSANOW, old_attributes)

        def _read_terminal_response(self, terminator: str) -> str:
            sys.stdin.read(1)
            sys.stdin.read(1)
            result = ""
            while True:
                ch = sys.stdin.read(1)
                if ch == terminator:
                    break
                result += ch
            return result

    def compare_applications(a: _frida.Application, b: _frida.Application) -> int:
        a_is_running = a.pid != 0
        b_is_running = b.pid != 0
        if a_is_running == b_is_running:
            if a.name > b.name:
                return 1
            elif a.name < b.name:
                return -1
            else:
                return 0
        elif a_is_running:
            return -1
        else:
            return 1

    def compare_processes(a: _frida.Process, b: _frida.Process) -> int:
        a_has_icon = "icons" in a.parameters
        b_has_icon = "icons" in b.parameters
        if a_has_icon == b_has_icon:
            if a.name > b.name:
                return 1
            elif a.name < b.name:
                return -1
            else:
                return 0
        elif a_has_icon:
            return -1
        else:
            return 1

    def compute_icon_width(item: Union[_frida.Application, _frida.Process]) -> int:
        for icon in item.parameters.get("icons", []):
            if icon["format"] == "png":
                return 4
        return 0

    app = PSApplication()
    app.run()


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



import argparse
import codecs
import os
import sys
from typing import Any, List

from colorama import Fore, Style

from frida_tools.application import ConsoleApplication


def main() -> None:
    app = RmApplication()
    app.run()


class RmApplication(ConsoleApplication):
    def _add_options(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("files", help="files to remove", nargs="+")
        parser.add_argument("-f", "--force", help="ignore nonexistent files", action="store_true")
        parser.add_argument(
            "-r", "--recursive", help="remove directories and their contents recursively", action="store_true"
        )

    def _usage(self) -> str:
        return "%(prog)s [options] FILE..."

    def _initialize(self, parser: argparse.ArgumentParser, options: argparse.Namespace, args: List[str]) -> None:
        self._paths = options.files
        self._flags = []
        if options.force:
            self._flags.append("force")
        if options.recursive:
            self._flags.append("recursive")

    def _needs_target(self) -> bool:
        return False

    def _start(self) -> None:
        try:
            self._attach(self._pick_worker_pid())

            data_dir = os.path.dirname(__file__)
            with codecs.open(os.path.join(data_dir, "fs_agent.js"), "r", "utf-8") as f:
                source = f.read()

            def on_message(message: Any, data: Any) -> None:
                self._reactor.schedule(lambda: self._on_message(message, data))

            assert self._session is not None
            script = self._session.create_script(name="pull", source=source)
            script.on("message", on_message)
            self._on_script_created(script)
            script.load()

            errors = script.exports_sync.rm(self._paths, self._flags)

            for message in errors:
                self._print(Fore.RED + Style.BRIGHT + message + Style.RESET_ALL, file=sys.stderr)

            status = 0 if len(errors) == 0 else 1
            self._exit(status)
        except Exception as e:
            self._update_status(str(e))
            self._exit(1)
            return

    def _on_message(self, message: Any, data: Any) -> None:
        print(message)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



import argparse
from typing import Any, List, MutableMapping


def main() -> None:
    from frida_tools.application import ConsoleApplication, await_ctrl_c

    class JoinApplication(ConsoleApplication):
        def __init__(self) -> None:
            ConsoleApplication.__init__(self, await_ctrl_c)
            self._parsed_options: MutableMapping[str, Any] = {}

        def _usage(self) -> str:
            return "%(prog)s [options] target portal-location [portal-certificate] [portal-token]"

        def _add_options(self, parser: argparse.ArgumentParser) -> None:
            parser.add_argument(
                "--portal-location", help="join portal at LOCATION", metavar="LOCATION", dest="portal_location"
            )
            parser.add_argument(
                "--portal-certificate",
                help="speak TLS with portal, expecting CERTIFICATE",
                metavar="CERTIFICATE",
                dest="portal_certificate",
            )
            parser.add_argument(
                "--portal-token", help="authenticate with portal using TOKEN", metavar="TOKEN", dest="portal_token"
            )
            parser.add_argument(
                "--portal-acl-allow",
                help="limit portal access to control channels with TAG",
                metavar="TAG",
                action="append",
                dest="portal_acl",
            )

        def _initialize(self, parser: argparse.ArgumentParser, options: argparse.Namespace, args: List[str]) -> None:
            location = args[0] if len(args) >= 1 else options.portal_location
            certificate = args[1] if len(args) >= 2 else options.portal_certificate
            token = args[2] if len(args) >= 3 else options.portal_token
            acl = options.portal_acl

            if location is None:
                parser.error("portal location must be specified")

            if certificate is not None:
                self._parsed_options["certificate"] = certificate
            if token is not None:
                self._parsed_options["token"] = token
            if acl is not None:
                self._parsed_options["acl"] = acl

            self._location = location

        def _needs_target(self) -> bool:
            return True

        def _start(self) -> None:
            self._update_status("Joining portal...")
            try:
                assert self._session is not None
                self._session.join_portal(self._location, **self._parsed_options)
            except Exception as e:
                self._update_status("Unable to join: " + str(e))
                self._exit(1)
                return
            self._update_status("Joined!")
            self._exit(0)

        def _stop(self) -> None:
            pass

    app = JoinApplication()
    app.run()


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



import argparse
import codecs
import os
import sys
import time
from threading import Event, Thread
from typing import AnyStr, List, MutableMapping, Optional

import frida
from colorama import Fore, Style

from frida_tools.application import ConsoleApplication
from frida_tools.stream_controller import DisposedException, StreamController
from frida_tools.units import bytes_to_megabytes


def main() -> None:
    app = PushApplication()
    app.run()


class PushApplication(ConsoleApplication):
    def _add_options(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("files", help="local files to push", nargs="+")

    def _usage(self) -> str:
        return "%(prog)s [options] LOCAL... REMOTE"

    def _initialize(self, parser: argparse.ArgumentParser, options: argparse.Namespace, args: List[str]) -> None:
        paths = options.files
        if len(paths) == 1:
            raise ValueError("missing remote path")
        self._local_paths = paths[:-1]
        self._remote_path = paths[-1]

        self._script: Optional[frida.core.Script] = None
        self._stream_controller: Optional[StreamController] = None
        self._total_bytes = 0
        self._time_started: Optional[float] = None
        self._completed = Event()
        self._transfers: MutableMapping[str, bool] = {}

    def _needs_target(self) -> bool:
        return False

    def _start(self) -> None:
        try:
            self._attach(self._pick_worker_pid())

            data_dir = os.path.dirname(__file__)
            with codecs.open(os.path.join(data_dir, "fs_agent.js"), "r", "utf-8") as f:
                source = f.read()

            def on_message(message, data) -> None:
                self._reactor.schedule(lambda: self._on_message(message, data))

            assert self._session is not None
            script = self._session.create_script(name="push", source=source)
            self._script = script
            script.on("message", on_message)
            self._on_script_created(script)
            script.load()

            self._stream_controller = StreamController(
                self._post_stream_stanza, on_stats_updated=self._on_stream_stats_updated
            )

            worker = Thread(target=self._perform_push)
            worker.start()
        except Exception as e:
            self._update_status(f"Failed to push: {e}")
            self._exit(1)
            return

    def _stop(self) -> None:
        for path in self._local_paths:
            if path not in self._transfers:
                self._complete_transfer(path, success=False)

        if self._stream_controller is not None:
            self._stream_controller.dispose()

    def _perform_push(self) -> None:
        for path in self._local_paths:
            try:
                self._total_bytes += os.path.getsize(path)
            except:
                pass
        self._time_started = time.time()

        for i, path in enumerate(self._local_paths):
            filename = os.path.basename(path)

            try:
                with open(path, "rb") as f:
                    assert self._stream_controller is not None
                    sink = self._stream_controller.open(str(i), {"filename": filename, "target": self._remote_path})
                    while True:
                        chunk = f.read(4 * 1024 * 1024)
                        if len(chunk) == 0:
                            break
                        sink.write(chunk)
                    sink.close()
            except DisposedException:
                break
            except Exception as e:
                self._print_error(str(e))
                self._complete_transfer(path, success=False)

        self._completed.wait()

        self._reactor.schedule(lambda: self._on_push_finished())

    def _on_push_finished(self) -> None:
        successes = self._transfers.values()

        if any(successes):
            self._render_summary_ui()

        status = 0 if all(successes) else 1
        self._exit(status)

    def _render_progress_ui(self) -> None:
        if self._completed.is_set():
            return
        assert self._stream_controller is not None
        megabytes_sent = bytes_to_megabytes(self._stream_controller.bytes_sent)
        total_megabytes = bytes_to_megabytes(self._total_bytes)
        if total_megabytes != 0 and megabytes_sent <= total_megabytes:
            self._update_status(f"Pushed {megabytes_sent:.1f} out of {total_megabytes:.1f} MB")
        else:
            self._update_status(f"Pushed {megabytes_sent:.1f} MB")

    def _render_summary_ui(self) -> None:
        assert self._time_started is not None
        duration = time.time() - self._time_started

        if len(self._local_paths) == 1:
            prefix = f"{self._local_paths[0]}: "
        else:
            prefix = ""

        files_transferred = sum(map(int, self._transfers.values()))

        assert self._stream_controller is not None
        bytes_sent = self._stream_controller.bytes_sent
        megabytes_per_second = bytes_to_megabytes(bytes_sent) / duration

        self._update_status(
            "{}{} file{} pushed. {:.1f} MB/s ({} bytes in {:.3f}s)".format(
                prefix,
                files_transferred,
                "s" if files_transferred != 1 else "",
                megabytes_per_second,
                bytes_sent,
                duration,
            )
        )

    def _on_message(self, message, data) -> None:
        handled = False

        if message["type"] == "send":
            payload = message["payload"]
            ptype = payload["type"]
            if ptype == "stream":
                stanza = payload["payload"]
                self._stream_controller.receive(stanza, data)
                handled = True
            elif ptype == "push:io-success":
                index = payload["index"]
                self._on_io_success(self._local_paths[index])
                handled = True
            elif ptype == "push:io-error":
                index = payload["index"]
                self._on_io_error(self._local_paths[index], payload["error"])
                handled = True

        if not handled:
            self._print(message)

    def _on_io_success(self, local_path: str) -> None:
        self._complete_transfer(local_path, success=True)

    def _on_io_error(self, local_path: str, error) -> None:
        self._print_error(f"{local_path}: {error}")
        self._complete_transfer(local_path, success=False)

    def _complete_transfer(self, local_path: str, success: bool) -> None:
        self._transfers[local_path] = success
        if len(self._transfers) == len(self._local_paths):
            self._completed.set()

    def _post_stream_stanza(self, stanza, data: Optional[AnyStr] = None) -> None:
        self._script.post({"type": "stream", "payload": stanza}, data=data)

    def _on_stream_stats_updated(self) -> None:
        self._render_progress_ui()

    def _print_error(self, message: str) -> None:
        self._print(Fore.RED + Style.BRIGHT + message + Style.RESET_ALL, file=sys.stderr)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



import argparse
import codecs
import os
import platform
from typing import Dict, List, Tuple

import frida

from frida_tools.application import ConsoleApplication


def main() -> None:
    app = CreatorApplication()
    app.run()


class CreatorApplication(ConsoleApplication):
    def _usage(self) -> str:
        return "%(prog)s [options] -t agent|cmodule"

    def _add_options(self, parser: argparse.ArgumentParser) -> None:
        default_project_name = os.path.basename(os.getcwd())
        parser.add_argument(
            "-n", "--project-name", help="project name", dest="project_name", default=default_project_name
        )
        parser.add_argument("-o", "--output-directory", help="output directory", dest="outdir", default=".")
        parser.add_argument("-t", "--template", help="template file: cmodule|agent", dest="template", default=None)

    def _initialize(self, parser: argparse.ArgumentParser, options: argparse.Namespace, args: List[str]) -> None:
        parsed_args = parser.parse_args()
        if not parsed_args.template:
            parser.error("template must be specified")
        impl = getattr(self, "_generate_" + parsed_args.template, None)
        if impl is None:
            parser.error("unknown template type")
        self._generate = impl

        self._project_name = options.project_name
        self._outdir = options.outdir

    def _needs_device(self) -> bool:
        return False

    def _start(self) -> None:
        (assets, message) = self._generate()

        outdir = self._outdir
        for name, data in assets.items():
            asset_path = os.path.join(outdir, name)

            asset_dir = os.path.dirname(asset_path)
            try:
                os.makedirs(asset_dir)
            except:
                pass

            with codecs.open(asset_path, "wb", "utf-8") as f:
                f.write(data)

            self._print("Created", asset_path)

        self._print("\n" + message)

        self._exit(0)

    def _generate_agent(self) -> Tuple[Dict[str, str], str]:
        assets = {}

        assets[
            "package.json"
        ] = f"""{{
  "name": "{self._project_name}-agent",
  "version": "1.0.0",
  "description": "Frida agent written in TypeScript",
  "private": true,
  "main": "agent/index.ts",
  "scripts": {{
    "prepare": "npm run build",
    "build": "frida-compile agent/index.ts -o _agent.js -c",
    "watch": "frida-compile agent/index.ts -o _agent.js -w"
  }},
  "devDependencies": {{
    "@types/frida-gum": "^18.3.1",
    "@types/node": "^18.14.0",
    "frida-compile": "^16.1.8"
  }}
}}
"""

        assets[
            "tsconfig.json"
        ] = """\
{
  "compilerOptions": {
    "target": "es2020",
    "lib": ["es2020"],
    "allowJs": true,
    "noEmit": true,
    "strict": true,
    "esModuleInterop": true,
    "moduleResolution": "node16"
  },
  "exclude": ["_agent.js"]
}
"""

        assets[
            "agent/index.ts"
        ] = """\
import { log } from "./logger.js";

const header = Memory.alloc(16);
header
    .writeU32(0xdeadbeef).add(4)
    .writeU32(0xd00ff00d).add(4)
    .writeU64(uint64("0x1122334455667788"));
log(hexdump(header.readByteArray(16) as ArrayBuffer, { ansi: true }));

Process.getModuleByName("libSystem.B.dylib")
    .enumerateExports()
    .slice(0, 16)
    .forEach((exp, index) => {
        log(`export ${index}: ${exp.name}`);
    });

Interceptor.attach(Module.getExportByName(null, "open"), {
    onEnter(args) {
        const path = args[0].readUtf8String();
        log(`open() path="${path}"`);
    }
});
"""

        assets[
            "agent/logger.ts"
        ] = """\
export function log(message: string): void {
    console.log(message);
}
"""

        assets[".gitignore"] = "/node_modules/\n"

        message = """\
Run `npm install` to bootstrap, then:
- Keep one terminal running: npm run watch
- Inject agent using the REPL: frida Calculator -l _agent.js
- Edit agent/*.ts - REPL will live-reload on save

Tip: Use an editor like Visual Studio Code for code completion, inline docs,
     instant type-checking feedback, refactoring tools, etc.
"""

        return (assets, message)

    def _generate_cmodule(self) -> Tuple[Dict[str, str], str]:
        assets = {}

        assets[
            "meson.build"
        ] = f"""\
project('{self._project_name}', 'c',
  default_options: 'buildtype=release',
)

shared_module('{self._project_name}', '{self._project_name}.c',
  name_prefix: '',
  include_directories: include_directories('include'),
)
"""

        assets[
            self._project_name + ".c"
        ] = """\
#include <gum/guminterceptor.h>

static void frida_log (const char * format, ...);
extern void _frida_log (const gchar * message);

void
init (void)
{
  frida_log ("init()");
}

void
finalize (void)
{
  frida_log ("finalize()");
}

void
on_enter (GumInvocationContext * ic)
{
  gpointer arg0;

  arg0 = gum_invocation_context_get_nth_argument (ic, 0);

  frida_log ("on_enter() arg0=%p", arg0);
}

void
on_leave (GumInvocationContext * ic)
{
  gpointer retval;

  retval = gum_invocation_context_get_return_value (ic);

  frida_log ("on_leave() retval=%p", retval);
}

static void
frida_log (const char * format,
           ...)
{
  gchar * message;
  va_list args;

  va_start (args, format);
  message = g_strdup_vprintf (format, args);
  va_end (args);

  _frida_log (message);

  g_free (message);
}
"""

        assets[".gitignore"] = "/build/\n"

        session = frida.attach(0)
        script = session.create_script("rpc.exports.getBuiltins = () => CModule.builtins;")
        self._on_script_created(script)
        script.load()
        builtins = script.exports_sync.get_builtins()
        script.unload()
        session.detach()

        for name, data in builtins["headers"].items():
            assets["include/" + name] = data

        system = platform.system()
        if system == "Windows":
            module_extension = "dll"
        elif system == "Darwin":
            module_extension = "dylib"
        else:
            module_extension = "so"

        cmodule_path = os.path.join(self._outdir, "build", self._project_name + "." + module_extension)

        message = f"""\
Run `meson build && ninja -C build` to build, then:
- Inject CModule using the REPL: frida Calculator -C {cmodule_path}
- Edit *.c, and build incrementally through `ninja -C build`
- REPL will live-reload whenever {cmodule_path} changes on disk
"""

        return (assets, message)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



import json
import os
import struct
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional, Sequence, Tuple, TypeVar, Union

import frida
from frida.core import RPCException

from frida_tools.reactor import Reactor

CodeLocation = Union[
    Tuple[str, str],
    Tuple[str, Tuple[str, str]],
    Tuple[str, Tuple[str, int]],
]

TraceThreadStrategy = Tuple[str, Tuple[str, int]]
TraceRangeStrategy = Tuple[str, Tuple[CodeLocation, Optional[CodeLocation]]]
TraceStrategy = Union[TraceThreadStrategy, TraceRangeStrategy]


def main() -> None:
    import argparse
    import threading

    from prompt_toolkit import PromptSession, prompt
    from prompt_toolkit.application import Application
    from prompt_toolkit.formatted_text import AnyFormattedText, FormattedText
    from prompt_toolkit.key_binding.defaults import load_key_bindings
    from prompt_toolkit.key_binding.key_bindings import KeyBindings, merge_key_bindings
    from prompt_toolkit.layout import Layout
    from prompt_toolkit.layout.containers import HSplit
    from prompt_toolkit.styles import BaseStyle
    from prompt_toolkit.widgets import Label, RadioList

    from frida_tools.application import ConsoleApplication

    class InstructionTracerApplication(ConsoleApplication, InstructionTracerUI):
        _itracer: Optional[InstructionTracer]

        def __init__(self) -> None:
            self._state = "starting"
            self._ready = threading.Event()
            self._cli = PromptSession()
            super().__init__(self._process_input)

        def _add_options(self, parser: argparse.ArgumentParser) -> None:
            parser.add_argument(
                "-t", "--thread-id", help="trace THREAD_ID", metavar="THREAD_ID", dest="strategy", type=parse_thread_id
            )
            parser.add_argument(
                "-i",
                "--thread-index",
                help="trace THREAD_INDEX",
                metavar="THREAD_INDEX",
                dest="strategy",
                type=parse_thread_index,
            )
            parser.add_argument(
                "-r",
                "--range",
                help="trace RANGE, e.g.: 0x1000..0x1008, libc.so!sleep, libc.so!0x1234, recv..memcpy",
                metavar="RANGE",
                dest="strategy",
                type=parse_range,
            )
            parser.add_argument("-o", "--output", help="output to file", dest="outpath")

        def _initialize(self, parser: argparse.ArgumentParser, options: argparse.Namespace, args: List[str]) -> None:
            self._itracer = None
            self._strategy = options.strategy
            self._outpath = options.outpath

        def _usage(self) -> str:
            return "%(prog)s [options] target"

        def _needs_target(self) -> bool:
            return True

        def _start(self) -> None:
            self._update_status("Injecting script...")
            self._itracer = InstructionTracer(self._reactor)
            self._itracer.start(self._device, self._session, self._runtime, self)
            self._ready.set()

        def _stop(self) -> None:
            assert self._itracer is not None
            self._itracer.dispose()
            self._itracer = None

            try:
                self._cli.app.exit()
            except:
                pass

        def _process_input(self, reactor: Reactor) -> None:
            try:
                while self._ready.wait(0.5) != True:
                    if not reactor.is_running():
                        return
            except KeyboardInterrupt:
                reactor.cancel_io()
                return

            if self._state != "started":
                return

            try:
                self._cli.prompt()
            except:
                pass

        def get_trace_strategy(self) -> Optional[TraceStrategy]:
            return self._strategy

        def prompt_for_trace_strategy(self, threads: List[dict]) -> Optional[TraceStrategy]:
            kind = radiolist_prompt(
                title="Tracing strategy:",
                values=[
                    ("thread", "Thread"),
                    ("range", "Range"),
                ],
            )
            if kind is None:
                raise KeyboardInterrupt

            if kind == "thread":
                thread_id = radiolist_prompt(
                    title="Running threads:", values=[(t["id"], json.dumps(t)) for t in threads]
                )
                if thread_id is None:
                    raise KeyboardInterrupt
                return ("thread", ("id", thread_id))

            while True:
                try:
                    text = prompt("Start address: ").strip()
                    if len(text) == 0:
                        continue
                    start = parse_code_location(text)
                    break
                except Exception as e:
                    print(str(e))
                    continue

            while True:
                try:
                    text = prompt("End address (optional): ").strip()
                    if len(text) > 0:
                        end = parse_code_location(text)
                    else:
                        end = None
                    break
                except Exception as e:
                    print(str(e))
                    continue

            return ("range", (start, end))

        def get_trace_output_path(self, suggested_name: Optional[str] = None) -> os.PathLike:
            return self._outpath

        def prompt_for_trace_output_path(self, suggested_name: str) -> Optional[os.PathLike]:
            while True:
                outpath = prompt("Output filename: ", default=suggested_name).strip()
                if len(outpath) != 0:
                    break
            return outpath

        def on_trace_started(self) -> None:
            self._state = "started"

        def on_trace_stopped(self, error_message: Optional[str] = None) -> None:
            self._state = "stopping"

            if error_message is not None:
                self._log(level="error", text=error_message)
                self._exit(1)
            else:
                self._exit(0)

            try:
                self._cli.app.exit()
            except:
                pass

        def on_trace_progress(self, total_blocks: int, total_bytes: int) -> None:
            blocks_suffix = "s" if total_blocks != 1 else ""
            self._cli.message = FormattedText(
                [
                    ("bold", "Tracing!"),
                    ("", " Collected "),
                    ("fg:green bold", human_readable_size(total_bytes)),
                    ("", f" from {total_blocks} basic block{blocks_suffix}"),
                ]
            )
            self._cli.app.invalidate()

    def parse_thread_id(value: str) -> TraceThreadStrategy:
        return ("thread", ("id", int(value)))

    def parse_thread_index(value: str) -> TraceThreadStrategy:
        return ("thread", ("index", int(value)))

    def parse_range(value: str) -> TraceRangeStrategy:
        tokens = value.split("..", 1)
        start = tokens[0]
        end = tokens[1] if len(tokens) == 2 else None
        return ("range", (parse_code_location(start), parse_code_location(end)))

    def parse_code_location(value: Optional[str]) -> CodeLocation:
        if value is None:
            return None

        if value.startswith("0x"):
            return ("address", value)

        tokens = value.split("!", 1)
        if len(tokens) == 2:
            name = tokens[0]
            subval = tokens[1]
            if subval.startswith("0x"):
                return ("module-offset", (name, int(subval, 16)))
            return ("module-export", (name, subval))

        return ("symbol", tokens[0])

    # Based on https://stackoverflow.com/a/43690506
    def human_readable_size(size):
        for unit in ["B", "KiB", "MiB", "GiB"]:
            if size < 1024.0 or unit == "GiB":
                break
            size /= 1024.0
        return f"{size:.2f} {unit}"

    T = TypeVar("T")

    # Based on https://github.com/prompt-toolkit/python-prompt-toolkit/issues/756#issuecomment-1294742392
    def radiolist_prompt(
        title: str = "",
        values: Sequence[Tuple[T, AnyFormattedText]] = None,
        default: Optional[T] = None,
        cancel_value: Optional[T] = None,
        style: Optional[BaseStyle] = None,
    ) -> T:
        radio_list = RadioList(values, default)
        radio_list.control.key_bindings.remove("enter")

        bindings = KeyBindings()

        @bindings.add("enter")
        def exit_with_value(event):
            radio_list._handle_enter()
            event.app.exit(result=radio_list.current_value)

        @bindings.add("c-c")
        def backup_exit_with_value(event):
            event.app.exit(result=cancel_value)

        application = Application(
            layout=Layout(HSplit([Label(title), radio_list])),
            key_bindings=merge_key_bindings([load_key_bindings(), bindings]),
            mouse_support=True,
            style=style,
            full_screen=False,
        )
        return application.run()

    app = InstructionTracerApplication()
    app.run()


class InstructionTracerUI(ABC):
    @abstractmethod
    def get_trace_strategy(self) -> Optional[TraceStrategy]:
        raise NotImplementedError

    def prompt_for_trace_strategy(self, threads: List[dict]) -> Optional[TraceStrategy]:
        return None

    @abstractmethod
    def get_trace_output_path(self) -> Optional[os.PathLike]:
        raise NotImplementedError

    def prompt_for_trace_output_path(self, suggested_name: str) -> Optional[os.PathLike]:
        return None

    @abstractmethod
    def on_trace_started(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def on_trace_stopped(self, error_message: Optional[str] = None) -> None:
        raise NotImplementedError

    def on_trace_progress(self, total_blocks: int, total_bytes: int) -> None:
        pass

    def _on_script_created(self, script: frida.core.Script) -> None:
        pass


class InstructionTracer:
    FILE_MAGIC = b"ITRC"

    def __init__(self, reactor: Reactor) -> None:
        self._reactor = reactor
        self._outfile = None
        self._ui: Optional[InstructionTracerUI] = None
        self._total_blocks = 0
        self._tracer_script: Optional[frida.core.Script] = None
        self._reader_script: Optional[frida.core.Script] = None
        self._reader_api = None

    def dispose(self) -> None:
        if self._reader_api is not None:
            try:
                self._reader_api.stop_buffer_reader()
            except:
                pass
            self._reader_api = None

        if self._reader_script is not None:
            try:
                self._reader_script.unload()
            except:
                pass
            self._reader_script = None

        if self._tracer_script is not None:
            try:
                self._tracer_script.unload()
            except:
                pass
            self._tracer_script = None

    def start(
        self, device: frida.core.Device, session: frida.core.Session, runtime: str, ui: InstructionTracerUI
    ) -> None:
        def on_message(message, data) -> None:
            self._reactor.schedule(lambda: self._on_message(message, data))

        self._ui = ui

        agent_source = (Path(__file__).parent / "itracer_agent.js").read_text(encoding="utf-8")

        try:
            tracer_script = session.create_script(name="itracer", source=agent_source, runtime=runtime)
            self._tracer_script = tracer_script
            self._ui._on_script_created(tracer_script)
            tracer_script.on("message", on_message)
            tracer_script.load()

            tracer_api = tracer_script.exports_sync

            outpath = ui.get_trace_output_path()
            if outpath is None:
                outpath = ui.prompt_for_trace_output_path(suggested_name=tracer_api.query_program_name() + ".itrace")
                if outpath is None:
                    ui.on_trace_stopped("Missing output path")
                    return

            self._outfile = open(outpath, "wb")
            self._outfile.write(self.FILE_MAGIC)

            strategy = ui.get_trace_strategy()
            if strategy is None:
                strategy = ui.prompt_for_trace_strategy(threads=tracer_api.list_threads())
                if strategy is None:
                    ui.on_trace_stopped("Missing strategy")
                    return

            buffer_location = tracer_api.create_buffer()

            try:
                system_session = device.attach(0)

                reader_script = system_session.create_script(name="itracer", source=agent_source, runtime=runtime)
                self._reader_script = reader_script
                self._ui._on_script_created(reader_script)
                reader_script.on("message", on_message)
                reader_script.load()

                reader_script.exports_sync.open_buffer(buffer_location)
            except:
                if self._reader_script is not None:
                    self._reader_script.unload()
                    self._reader_script = None
                reader_script = None

            if reader_script is not None:
                reader_api = reader_script.exports_sync
            else:
                reader_api = tracer_script.exports_sync
            self._reader_api = reader_api
            reader_api.launch_buffer_reader()

            tracer_script.exports_sync.launch_trace_session(strategy)

            ui.on_trace_started()
        except RPCException as e:
            ui.on_trace_stopped(f"Unable to start: {e.args[0]}")
        except Exception as e:
            ui.on_trace_stopped(str(e))
        except KeyboardInterrupt:
            ui.on_trace_stopped()

    def _on_message(self, message, data) -> None:
        handled = False

        if message["type"] == "send":
            try:
                payload = message["payload"]
                mtype = payload["type"]
                params = (mtype, payload, data)
            except:
                params = None
            if params is not None:
                handled = self._try_handle_message(*params)

        if not handled:
            print(message)

    def _try_handle_message(self, mtype, message, data) -> bool:
        if not mtype.startswith("itrace:"):
            return False

        if mtype == "itrace:chunk":
            self._write_chunk(data)
        else:
            self._write_message(message, data)

            if mtype == "itrace:compile":
                self._total_blocks += 1

        self._update_progress()

        if mtype == "itrace:end":
            self._ui.on_trace_stopped()

        return True

    def _update_progress(self) -> None:
        self._ui.on_trace_progress(self._total_blocks, self._outfile.tell())

    def _write_message(self, message, data) -> None:
        f = self._outfile

        raw_message = json.dumps(message).encode("utf-8")
        f.write(struct.pack(">II", RecordType.MESSAGE, len(raw_message)))
        f.write(raw_message)

        data_size = len(data) if data is not None else 0
        f.write(struct.pack(">I", data_size))
        if data_size != 0:
            f.write(data)

        f.flush()

    def _write_chunk(self, chunk) -> None:
        f = self._outfile
        f.write(struct.pack(">II", RecordType.CHUNK, len(chunk)))
        f.write(chunk)
        f.flush()


class RecordType:
    MESSAGE = 1
    CHUNK = 2


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



from __future__ import annotations

import argparse
import asyncio
import binascii
import codecs
import email.utils
import errno
import gzip
import http
import mimetypes
import os
import re
import shlex
import subprocess
import threading
from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Generator, List, Optional, Set
from zipfile import ZipFile

import frida
import websockets.asyncio.server
import websockets.datastructures
import websockets.exceptions
import websockets.http11

from frida_tools.reactor import Reactor

MANPAGE_CONTROL_CHARS = re.compile(r"\.[a-zA-Z]*(\s|$)|\s?\"")
MANPAGE_FUNCTION_PROTOTYPE = re.compile(r"([a-zA-Z_]\w+)\(([^\)]+)")


def main() -> None:
    import json

    from colorama import Fore, Style

    from frida_tools.application import ConsoleApplication, await_ctrl_c

    class TracerApplication(ConsoleApplication, UI):
        def __init__(self) -> None:
            super().__init__(await_ctrl_c)
            self._handlers = OrderedDict()
            self._ui_port = 1337
            self._ui_zip = ZipFile(Path(__file__).parent / "tracer_ui.zip", "r")
            self._ui_sockets: Set[websockets.asyncio.server.ServerConnection] = set()
            self._ui_worker = None
            self._asyncio_loop = None
            self._palette = ["cyan", "magenta", "yellow", "green", "red", "blue"]
            self._next_color = 0
            self._style_by_thread_id = {}
            self._last_event_tid = -1

        def _add_options(self, parser: argparse.ArgumentParser) -> None:
            pb = TracerProfileBuilder()
            parser.add_argument(
                "-I", "--include-module", help="include MODULE", metavar="MODULE", type=pb.include_modules
            )
            parser.add_argument(
                "-X", "--exclude-module", help="exclude MODULE", metavar="MODULE", type=pb.exclude_modules
            )
            parser.add_argument(
                "-i", "--include", help="include [MODULE!]FUNCTION", metavar="FUNCTION", type=pb.include
            )
            parser.add_argument(
                "-x", "--exclude", help="exclude [MODULE!]FUNCTION", metavar="FUNCTION", type=pb.exclude
            )
            parser.add_argument(
                "-a", "--add", help="add MODULE!OFFSET", metavar="MODULE!OFFSET", type=pb.include_relative_address
            )
            parser.add_argument("-T", "--include-imports", help="include program's imports", type=pb.include_imports)
            parser.add_argument(
                "-t",
                "--include-module-imports",
                help="include MODULE imports",
                metavar="MODULE",
                type=pb.include_imports,
            )
            parser.add_argument(
                "-m",
                "--include-objc-method",
                help="include OBJC_METHOD",
                metavar="OBJC_METHOD",
                type=pb.include_objc_method,
            )
            parser.add_argument(
                "-M",
                "--exclude-objc-method",
                help="exclude OBJC_METHOD",
                metavar="OBJC_METHOD",
                type=pb.exclude_objc_method,
            )
            parser.add_argument(
                "-y",
                "--include-swift-func",
                help="include SWIFT_FUNC",
                metavar="SWIFT_FUNC",
                type=pb.include_swift_func,
            )
            parser.add_argument(
                "-Y",
                "--exclude-swift-func",
                help="exclude SWIFT_FUNC",
                metavar="SWIFT_FUNC",
                type=pb.exclude_swift_func,
            )
            parser.add_argument(
                "-j",
                "--include-java-method",
                help="include JAVA_METHOD",
                metavar="JAVA_METHOD",
                type=pb.include_java_method,
            )
            parser.add_argument(
                "-J",
                "--exclude-java-method",
                help="exclude JAVA_METHOD",
                metavar="JAVA_METHOD",
                type=pb.exclude_java_method,
            )
            parser.add_argument(
                "-s",
                "--include-debug-symbol",
                help="include DEBUG_SYMBOL",
                metavar="DEBUG_SYMBOL",
                type=pb.include_debug_symbol,
            )
            parser.add_argument(
                "-q", "--quiet", help="do not format output messages", action="store_true", default=False
            )
            parser.add_argument(
                "-d",
                "--decorate",
                help="add module name to generated onEnter log statement",
                action="store_true",
                default=False,
            )
            parser.add_argument(
                "-S",
                "--init-session",
                help="path to JavaScript file used to initialize the session",
                metavar="PATH",
                action="append",
                default=[],
            )
            parser.add_argument(
                "-P",
                "--parameters",
                help="parameters as JSON, exposed as a global named 'parameters'",
                metavar="PARAMETERS_JSON",
            )
            parser.add_argument("-o", "--output", help="dump messages to file", metavar="OUTPUT")
            self._profile_builder = pb

        def _usage(self) -> str:
            return "%(prog)s [options] target"

        def _initialize(self, parser: argparse.ArgumentParser, options: argparse.Namespace, args: List[str]) -> None:
            self._repo: Optional[FileRepository] = None
            self._tracer: Optional[Tracer] = None
            self._profile = self._profile_builder.build()
            self._quiet: bool = options.quiet
            self._decorate: bool = options.decorate
            self._output: Optional[OutputFile] = None
            self._output_path: str = options.output

            self._init_scripts = []
            for path in options.init_session:
                with codecs.open(path, "rb", "utf-8") as f:
                    source = f.read()
                self._init_scripts.append(InitScript(path, source))

            if options.parameters is not None:
                try:
                    params = json.loads(options.parameters)
                except Exception as e:
                    raise ValueError(f"failed to parse parameters argument as JSON: {e}")
                if not isinstance(params, dict):
                    raise ValueError("failed to parse parameters argument as JSON: not an object")
                self._parameters = params
            else:
                self._parameters = {}

        def _needs_target(self) -> bool:
            return True

        def _start(self) -> None:
            if self._ui_worker is None:
                worker = threading.Thread(target=self._run_ui_server, name="ui-server", daemon=True)
                worker.start()
                self._ui_worker = worker

            if self._output_path is not None:
                self._output = OutputFile(self._output_path)

            stage = "early" if self._target[0] == "file" else "late"

            self._repo = FileRepository(self._reactor, self._decorate)
            self._tracer = Tracer(
                self._reactor,
                self._repo,
                self._profile,
                self._init_scripts,
                log_handler=self._log,
            )
            try:
                self._tracer.start_trace(self._session, stage, self._parameters, self._runtime, self)
            except Exception as e:
                self._update_status(f"Failed to start tracing: {e}")
                self._exit(1)
                return

        def _stop(self) -> None:
            self._tracer.stop()
            self._tracer = None
            if self._output is not None:
                self._output.close()
            self._output = None

            self._handlers.clear()
            self._next_color = 0
            self._style_by_thread_id.clear()
            self._last_event_tid = -1

        def on_script_created(self, script: frida.core.Script) -> None:
            self._on_script_created(script)

        def on_trace_progress(self, status: str, *params) -> None:
            if status == "initializing":
                self._update_status("Instrumenting...")
            elif status == "initialized":
                self._resume()
            elif status == "started":
                (count,) = params
                if count == 1:
                    plural = ""
                else:
                    plural = "s"
                self._update_status(
                    f"Started tracing {count} function{plural}. Web UI available at http://localhost:{self._ui_port}/"
                )

        def on_trace_warning(self, message: str) -> None:
            self._print(Fore.RED + Style.BRIGHT + "Warning" + Style.RESET_ALL + ": " + message)

        def on_trace_error(self, message: str) -> None:
            self._print(Fore.RED + Style.BRIGHT + "Error" + Style.RESET_ALL + ": " + message)
            self._exit(1)

        def on_trace_events(self, raw_events) -> None:
            events = [
                (target_id, timestamp, thread_id, depth, message, self._get_style(thread_id))
                for target_id, timestamp, thread_id, depth, message in raw_events
            ]
            self._asyncio_loop.call_soon_threadsafe(
                lambda: self._asyncio_loop.create_task(self._broadcast_trace_events(events))
            )

            no_attributes = Style.RESET_ALL
            for target_id, timestamp, thread_id, depth, message, style in events:
                if self._output is not None:
                    self._output.append(message + "\n")
                elif self._quiet:
                    self._print(message)
                else:
                    indent = depth * "   | "
                    attributes = getattr(Fore, style[0].upper())
                    if len(style) > 1:
                        attributes += getattr(Style, style[1].upper())
                    if thread_id != self._last_event_tid:
                        self._print("%s           /* TID 0x%x */%s" % (attributes, thread_id, Style.RESET_ALL))
                        self._last_event_tid = thread_id
                    self._print("%6d ms  %s%s%s%s" % (timestamp, attributes, indent, message, no_attributes))

        def on_trace_handler_create(self, target: TraceTarget, handler: str, source: str) -> None:
            self._register_handler(target, source)
            if self._quiet:
                return
            self._print('%s: Auto-generated handler at "%s"' % (target, source.replace("\\", "\\\\")))

        def on_trace_handler_load(self, target: TraceTarget, handler: str, source: str) -> None:
            self._register_handler(target, source)
            if self._quiet:
                return
            self._print('%s: Loaded handler at "%s"' % (target, source.replace("\\", "\\\\")))

        def _register_handler(self, target: TraceTarget, source: str) -> None:
            self._handlers[target.identifier] = (target, source)

        def _get_style(self, thread_id):
            style = self._style_by_thread_id.get(thread_id, None)
            if style is None:
                color = self._next_color
                self._next_color += 1
                style = [self._palette[color % len(self._palette)]]
                if (1 + int(color / len(self._palette))) % 2 == 0:
                    style.append("bright")
                self._style_by_thread_id[thread_id] = style
            return style

        def _run_ui_server(self):
            asyncio.run(self._handle_ui_requests())

        async def _handle_ui_requests(self):
            self._asyncio_loop = asyncio.get_running_loop()
            while True:
                try:
                    async with websockets.asyncio.server.serve(
                        self._handle_websocket_connection,
                        "localhost",
                        self._ui_port,
                        process_request=self._handle_asset_request,
                    ):
                        await asyncio.get_running_loop().create_future()
                        return
                except OSError as e:
                    if e.errno == errno.EADDRINUSE:
                        self._ui_port += 1
                    else:
                        raise

        async def _handle_websocket_connection(self, websocket: websockets.asyncio.server.ServerConnection):
            if self._tracer is None:
                return

            self._ui_sockets.add(websocket)
            try:
                message = {
                    "type": "tracer:sync",
                    "spawned_program": self._spawned_argv[0] if self._spawned_argv is not None else None,
                    "handlers": [target.to_json() for target, source in self._handlers.values()],
                }
                await websocket.send(json.dumps(message))

                while True:
                    message = json.loads(await websocket.recv())
                    mtype = message["type"]
                    if mtype == "tracer:respawn":
                        self._reactor.schedule(self._respawn)
                    elif mtype == "handler:load":
                        target, source = self._handlers[message["id"]]
                        await websocket.send(
                            json.dumps(
                                {
                                    "type": "handler:loaded",
                                    "code": self._repo.ensure_handler(target),
                                }
                            )
                        )
                    elif mtype == "handler:save":
                        target, source = self._handlers[message["id"]]
                        self._repo.update_handler(target, message["code"])
                    elif mtype == "targets:stage":
                        profile = TracerProfile(list(map(tuple, message["profile"]["spec"])))
                        try:
                            items = self._tracer.stage_targets(profile)
                        except:
                            continue
                        await websocket.send(
                            json.dumps(
                                {
                                    "type": "targets:staged",
                                    "items": items,
                                }
                            )
                        )
                    elif mtype == "targets:commit":
                        target_ids = self._tracer.commit_targets(message["id"])
                        message = {
                            "type": "handlers:add",
                            "handlers": [self._handlers[target_id][0].to_json() for target_id in target_ids],
                        }
                        await websocket.send(json.dumps(message))
            except:
                pass
            finally:
                self._ui_sockets.remove(websocket)

        async def _broadcast_trace_events(self, events):
            for websocket in self._ui_sockets:
                await websocket.send(
                    json.dumps(
                        {
                            "type": "events:add",
                            "events": events,
                        }
                    )
                )

        def _handle_asset_request(
            self, connection: websockets.asyncio.server.ServerConnection, request: websockets.asyncio.server.Request
        ):
            if request.headers.get("Connection") == "Upgrade":
                return

            filename = request.path[1:]
            if filename == "":
                filename = "index.html"

            try:
                body = self._ui_zip.read(filename)
            except KeyError:
                return connection.respond(http.HTTPStatus.NOT_FOUND, "File not found\n")

            status = http.HTTPStatus(http.HTTPStatus.OK)

            content_type, content_encoding = mimetypes.guess_type(filename)
            if content_type is None:
                content_type = "application/octet-stream"

            headers = websockets.datastructures.Headers(
                [
                    ("Connection", "close"),
                    ("Content-Length", str(len(body))),
                    ("Content-Type", content_type),
                    ("Date", email.utils.formatdate(usegmt=True)),
                ]
            )
            if content_encoding is not None:
                headers.update({"Content-Encoding": content_encoding})

            response = websockets.http11.Response(status.value, status.phrase, headers, body)
            connection.protocol.handshake_exc = websockets.exceptions.InvalidStatus(response)

            return response

    app = TracerApplication()
    app.run()


class TracerProfileBuilder:
    def __init__(self) -> None:
        self._spec = []

    def include_modules(self, *module_name_globs: str) -> "TracerProfileBuilder":
        for m in module_name_globs:
            self._spec.append(("include", "module", m))
        return self

    def exclude_modules(self, *module_name_globs: str) -> "TracerProfileBuilder":
        for m in module_name_globs:
            self._spec.append(("exclude", "module", m))
        return self

    def include(self, *function_name_globs: str) -> "TracerProfileBuilder":
        for f in function_name_globs:
            self._spec.append(("include", "function", f))
        return self

    def exclude(self, *function_name_globs: str) -> "TracerProfileBuilder":
        for f in function_name_globs:
            self._spec.append(("exclude", "function", f))
        return self

    def include_relative_address(self, *address_rel_offsets: str) -> "TracerProfileBuilder":
        for f in address_rel_offsets:
            self._spec.append(("include", "relative-function", f))
        return self

    def include_imports(self, *module_name_globs: str) -> "TracerProfileBuilder":
        for m in module_name_globs:
            self._spec.append(("include", "imports", m))
        return self

    def include_objc_method(self, *function_name_globs: str) -> "TracerProfileBuilder":
        for f in function_name_globs:
            self._spec.append(("include", "objc-method", f))
        return self

    def exclude_objc_method(self, *function_name_globs: str) -> "TracerProfileBuilder":
        for f in function_name_globs:
            self._spec.append(("exclude", "objc-method", f))
        return self

    def include_swift_func(self, *function_name_globs: str) -> "TracerProfileBuilder":
        for f in function_name_globs:
            self._spec.append(("include", "swift-func", f))
        return self

    def exclude_swift_func(self, *function_name_globs: str) -> "TracerProfileBuilder":
        for f in function_name_globs:
            self._spec.append(("exclude", "swift-func", f))
        return self

    def include_java_method(self, *function_name_globs: str) -> "TracerProfileBuilder":
        for f in function_name_globs:
            self._spec.append(("include", "java-method", f))
        return self

    def exclude_java_method(self, *function_name_globs: str) -> "TracerProfileBuilder":
        for f in function_name_globs:
            self._spec.append(("exclude", "java-method", f))
        return self

    def include_debug_symbol(self, *function_name_globs: str) -> "TracerProfileBuilder":
        for f in function_name_globs:
            self._spec.append(("include", "debug-symbol", f))
        return self

    def build(self) -> "TracerProfile":
        return TracerProfile(self._spec)


class TracerProfile:
    def __init__(self, spec) -> None:
        self.spec = spec


class Tracer:
    def __init__(
        self,
        reactor: Reactor,
        repository: "Repository",
        profile: TracerProfile,
        init_scripts=[],
        log_handler: Callable[[str, str], None] = None,
    ) -> None:
        self._reactor = reactor
        self._repository = repository
        self._profile = profile
        self._script: Optional[frida.core.Script] = None
        self._schedule_on_message = None
        self._agent = None
        self._init_scripts = init_scripts
        self._log_handler = log_handler

    def start_trace(self, session: frida.core.Session, stage, parameters, runtime, ui: UI) -> None:
        def on_create(*args) -> None:
            ui.on_trace_handler_create(*args)

        self._repository.on_create(on_create)

        def on_load(*args) -> None:
            ui.on_trace_handler_load(*args)

        self._repository.on_load(on_load)

        def on_update(target, handler, source) -> None:
            self._agent.update(target.identifier, target.display_name, handler)

        self._repository.on_update(on_update)

        self._schedule_on_message = lambda message, data: self._reactor.schedule(
            lambda: self._on_message(message, data, ui)
        )

        ui.on_trace_progress("initializing")
        data_dir = os.path.dirname(__file__)
        with codecs.open(os.path.join(data_dir, "tracer_agent.js"), "r", "utf-8") as f:
            source = f.read()
        script = session.create_script(name="tracer", source=source, runtime=runtime)

        self._script = script
        script.set_log_handler(self._log_handler)
        script.on("message", self._schedule_on_message)
        ui.on_script_created(script)
        script.load()

        self._agent = script.exports_sync

        raw_init_scripts = [{"filename": script.filename, "source": script.source} for script in self._init_scripts]
        self._agent.init(stage, parameters, raw_init_scripts, self._profile.spec)

    def stop(self) -> None:
        self._repository.close()

        if self._script is not None:
            self._script.off("message", self._schedule_on_message)
            try:
                self._script.unload()
            except:
                pass
            self._script = None

    def stage_targets(self, profile: TracerProfile) -> List:
        return self._agent.stage_targets(profile.spec)

    def commit_targets(self, identifier: Optional[int]) -> List:
        return self._agent.commit_targets(identifier)

    def _on_message(self, message, data, ui) -> None:
        handled = False

        if message["type"] == "send":
            try:
                payload = message["payload"]
                mtype = payload["type"]
                params = (mtype, payload, data, ui)
            except:
                # As user scripts may use send() we need to be prepared for this.
                params = None
            if params is not None:
                handled = self._try_handle_message(*params)

        if not handled:
            print(message)

    def _try_handle_message(self, mtype, params, data, ui) -> False:
        if mtype == "events:add":
            events = [
                (target_id, timestamp, thread_id, depth, message)
                for target_id, timestamp, thread_id, depth, message in params["events"]
            ]
            ui.on_trace_events(events)
            return True

        if mtype == "handlers:get":
            flavor = params["flavor"]
            base_id = params["baseId"]

            scripts = []
            response = {"type": f"reply:{base_id}", "scripts": scripts}

            repo = self._repository
            next_id = base_id
            for scope in params["scopes"]:
                scope_name = scope["name"]
                for member_name in scope["members"]:
                    if isinstance(member_name, list):
                        name, display_name = member_name
                    else:
                        name = member_name
                        display_name = member_name
                    target = TraceTarget(next_id, flavor, scope_name, name, display_name)
                    next_id += 1
                    handler = repo.ensure_handler(target)
                    scripts.append(handler)

            self._script.post(response)

            return True

        if mtype == "agent:initialized":
            ui.on_trace_progress("initialized")
            return True

        if mtype == "agent:started":
            self._repository.commit_handlers()
            ui.on_trace_progress("started", params["count"])
            return True

        if mtype == "agent:warning":
            ui.on_trace_warning(params["message"])
            return True

        if mtype == "agent:error":
            ui.on_trace_error(params["message"])
            return True

        return False


@dataclass
class TraceTarget:
    identifier: int
    flavor: str
    scope: str
    name: str
    display_name: str

    def to_json(self) -> dict:
        return {"id": self.identifier, "scope": self.scope, "display_name": self.display_name}

    def __str__(self) -> str:
        return self.display_name


class Repository:
    def __init__(self) -> None:
        self._on_create_callback: Optional[Callable[[TraceTarget, str, str], None]] = None
        self._on_load_callback: Optional[Callable[[TraceTarget, str, str], None]] = None
        self._on_update_callback: Optional[Callable[[TraceTarget, str, str], None]] = None
        self._decorate = False
        self._manpages = None

    def close(self) -> None:
        self._on_create_callback = None
        self._on_load_callback = None
        self._on_update_callback = None

    def ensure_handler(self, target: TraceTarget):
        raise NotImplementedError("not implemented")

    def commit_handlers(self) -> None:
        pass

    def on_create(self, callback: Callable[[TraceTarget, str, str], None]) -> None:
        self._on_create_callback = callback

    def on_load(self, callback: Callable[[TraceTarget, str, str], None]) -> None:
        self._on_load_callback = callback

    def on_update(self, callback: Callable[[TraceTarget, str, str], None]) -> None:
        self._on_update_callback = callback

    def _notify_create(self, target: TraceTarget, handler: str, source: str) -> None:
        if self._on_create_callback is not None:
            self._on_create_callback(target, handler, source)

    def _notify_load(self, target: TraceTarget, handler: str, source: str) -> None:
        if self._on_load_callback is not None:
            self._on_load_callback(target, handler, source)

    def _notify_update(self, target: TraceTarget, handler: str, source: str) -> None:
        if self._on_update_callback is not None:
            self._on_update_callback(target, handler, source)

    def _create_stub_handler(self, target: TraceTarget, decorate: bool) -> str:
        if target.flavor == "java":
            return self._create_stub_java_handler(target, decorate)
        else:
            return self._create_stub_native_handler(target, decorate)

    def _create_stub_native_handler(self, target: TraceTarget, decorate: bool) -> str:
        if target.flavor == "objc":
            log_str = self._create_objc_logging_code(target)
        elif target.flavor == "swift":
            log_str = self._create_swift_logging_code(target, decorate)
        else:
            log_str = self._create_cstyle_logging_code(target, decorate)

        return """\
/*
 * Auto-generated by Frida. Please modify to match the signature of %(display_name)s.
 * This stub is currently auto-generated from manpages when available.
 *
 * For full API reference, see: https://frida.re/docs/javascript-api/
 */

defineHandler({
  onEnter(log, args, state) {
    log(%(log_str)s);
  },

  onLeave(log, retval, state) {
  }
});
""" % {
            "display_name": target.display_name,
            "log_str": log_str,
        }

    def _create_cstyle_logging_code(self, target: TraceTarget, decorate: bool) -> str:
        if decorate:
            module_string = " [%s]" % os.path.basename(target.scope)
        else:
            module_string = ""

        args = self._generate_cstyle_argument_logging_code(target)
        if len(args) == 0:
            code = "'%(name)s()%(module_string)s'" % {"name": target.name, "module_string": module_string}
        else:
            code = "`%(name)s(%(args)s)%(module_string)s`" % {
                "name": target.name,
                "args": ", ".join(args),
                "module_string": module_string,
            }

        return code

    def _create_objc_logging_code(self, target: TraceTarget) -> str:
        state = {"index": 2}

        def objc_arg(m):
            index = state["index"]
            r = ":${args[%d]} " % index
            state["index"] = index + 1
            return r

        code = "`" + re.sub(r":", objc_arg, target.display_name) + "`"
        if code.endswith("} ]`"):
            code = code[:-3] + "]`"

        return code

    def _create_swift_logging_code(self, target: TraceTarget, decorate: bool) -> str:
        if decorate:
            module_string = " [%s]" % os.path.basename(target.scope)
        else:
            module_string = ""
        return "'%(name)s()%(module_string)s'" % {"name": target.name, "module_string": module_string}

    def _create_stub_java_handler(self, target: TraceTarget, decorate) -> str:
        return """\
/*
 * Auto-generated by Frida. Please modify to match the signature of %(display_name)s.
 *
 * For full API reference, see: https://frida.re/docs/javascript-api/
 */

{
  /**
   * Called synchronously when about to call %(display_name)s.
   *
   * @this {object} - The Java class or instance.
   * @param {function} log - Call this function with a string to be presented to the user.
   * @param {array} args - Java method arguments.
   * @param {object} state - Object allowing you to keep state across function calls.
   */
  onEnter(log, args, state) {
    log(`%(display_name)s(${args.map(JSON.stringify).join(', ')})`);
  },

  /**
   * Called synchronously when about to return from %(display_name)s.
   *
   * See onEnter for details.
   *
   * @this {object} - The Java class or instance.
   * @param {function} log - Call this function with a string to be presented to the user.
   * @param {NativePointer} retval - Return value.
   * @param {object} state - Object allowing you to keep state across function calls.
   */
  onLeave(log, retval, state) {
    if (retval !== undefined) {
      log(`<= ${JSON.stringify(retval)}`);
    }
  }
}
""" % {
            "display_name": target.display_name
        }

    def _generate_cstyle_argument_logging_code(self, target: TraceTarget) -> List[str]:
        if self._manpages is None:
            self._manpages = {}
            try:
                manroots = [
                    Path(d)
                    for d in subprocess.run(["manpath"], stdout=subprocess.PIPE, encoding="utf-8", check=True)
                    .stdout.strip()
                    .split(":")
                ]
                for section in (2, 3):
                    for manroot in manroots:
                        mandir = manroot / f"man{section}"
                        if not mandir.exists():
                            continue
                        raw_section = str(section)
                        for entry in mandir.iterdir():
                            tokens = entry.name.split(".")
                            if len(tokens) < 2:
                                continue
                            if not tokens[1].startswith(raw_section):
                                continue
                            name = tokens[0]
                            if name in self._manpages:
                                continue
                            self._manpages[name] = (entry, section)
            except:
                return []

        man_entry = self._manpages.get(target.name)
        if man_entry is None:
            return []
        man_location, man_section = man_entry

        try:
            args = []
            cfunc = next(f for f in self._read_manpage(man_location) if f.name == target.name)
            for arg in cfunc.arguments:
                if arg == "void":
                    continue
                if arg.startswith("..."):
                    args.append("...")
                    continue

                tokens = arg.split(" ")

                arg_type = "".join(tokens[:-1])

                arg_name = tokens[-1]
                if arg_name.startswith("*"):
                    arg_type += "*"
                    arg_name = arg_name[1:]
                elif arg_name.endswith("]"):
                    arg_type += "*"
                    arg_name = arg_name[: arg_name.index("[")]

                read_ops = ""
                annotate_pre = ""
                annotate_post = ""

                if arg_type.endswith("*restrict"):
                    arg_type = arg_type[:-8]
                if arg_type in ("char*", "constchar*"):
                    read_ops = ".readUtf8String()"
                    annotate_pre = '"'
                    annotate_post = '"'

                arg_index = len(args)

                args.append(
                    "%(arg_name)s=%(annotate_pre)s${args[%(arg_index)s]%(read_ops)s}%(annotate_post)s"
                    % {
                        "arg_name": arg_name,
                        "arg_index": arg_index,
                        "read_ops": read_ops,
                        "annotate_pre": annotate_pre,
                        "annotate_post": annotate_post,
                    }
                )
            return args
        except Exception:
            return []

    def _read_manpage(self, man_location: Path) -> Generator[CFuncSpec]:
        if man_location.suffix == ".gz":
            man_file = gzip.open(man_location, "rt", encoding="utf-8", errors="replace")
        else:
            man_file = open(man_location, "r", encoding="utf-8", errors="replace")
        with man_file:
            man_data = man_file.read()

        manpage_format = "gnu"
        synopsis_lines = []
        found_synopsis = False
        in_multiline = False
        for raw_line in man_data.split("\n"):
            line = raw_line.strip()
            if line.startswith(".so "):
                redirected_location = man_location.parent.parent / Path(line[4:])
                if not redirected_location.exists():
                    redirected_location = redirected_location.parent / (redirected_location.name + ".gz")
                yield from self._read_manpage(redirected_location)
                return
            if not found_synopsis and "SYNOPSIS" in line:
                found_synopsis = True
                continue
            elif found_synopsis and line.endswith("DESCRIPTION"):
                break
            elif not found_synopsis:
                continue
            if line.startswith(".Fn ") or line.startswith(".Fo "):
                manpage_format = "bsd"
            escaped_newline = line.endswith("\\")
            if escaped_newline:
                line = line[:-1]
            if in_multiline:
                synopsis_lines[-1] += line
            else:
                synopsis_lines.append(line)
            in_multiline = escaped_newline

        if manpage_format == "gnu":
            raw_synopsis = "\n".join(synopsis_lines)
            synopsis = (
                MANPAGE_CONTROL_CHARS.sub("", raw_synopsis).replace("\n", " ").replace(" [", "[").replace(" ]", "]")
            )

            for m in MANPAGE_FUNCTION_PROTOTYPE.finditer(synopsis):
                name = m.group(1)
                signature = m.group(2)
                args = [a.strip() for a in signature.split(",")]
                yield CFuncSpec(name, args)
        else:
            name = None
            args = None
            for line in synopsis_lines:
                tokens = line.split(" ", maxsplit=1)
                directive = tokens[0]
                data = tokens[1] if len(tokens) == 2 else None
                if directive == ".Fn":
                    argv = shlex.split(data)
                    yield CFuncSpec(argv[0], argv[1:])
                elif directive == ".Fo":
                    name = data
                    args = []
                elif directive == ".Fa":
                    args.append(shlex.split(data)[0])
                elif directive == ".Fc":
                    yield CFuncSpec(name, args)


@dataclass
class CFuncSpec:
    name: str
    arguments: List[str]


class MemoryRepository(Repository):
    def __init__(self) -> None:
        super().__init__()
        self._handlers = {}

    def ensure_handler(self, target: TraceTarget) -> str:
        handler = self._handlers.get(target)
        if handler is None:
            handler = self._create_stub_handler(target, False)
            self._handlers[target] = handler
            self._notify_create(target, handler, "memory")
        else:
            self._notify_load(target, handler, "memory")
        return handler


class FileRepository(Repository):
    def __init__(self, reactor: Reactor, decorate: bool) -> None:
        super().__init__()
        self._reactor = reactor
        self._handler_by_id = {}
        self._handler_by_file = {}
        self._changed_files = set()
        self._last_change_id = 0
        self._repo_dir = os.path.join(os.getcwd(), "__handlers__")
        self._repo_monitors = {}
        self._decorate = decorate

    def close(self) -> None:
        for monitor in self._repo_monitors.values():
            monitor.disable()
        self._repo_monitors.clear()

        super().close()

    def ensure_handler(self, target: TraceTarget) -> str:
        entry = self._handler_by_id.get(target.identifier)
        if entry is not None:
            (target, handler, handler_file) = entry
            return handler

        handler = None

        scope = target.scope
        if len(scope) > 0:
            handler_file = os.path.join(
                self._repo_dir, to_filename(os.path.basename(scope)), to_handler_filename(target.name)
            )
        else:
            handler_file = os.path.join(self._repo_dir, to_handler_filename(target.name))

        if os.path.isfile(handler_file):
            with codecs.open(handler_file, "r", "utf-8") as f:
                handler = f.read()
            self._notify_load(target, handler, handler_file)

        if handler is None:
            handler = self._create_stub_handler(target, self._decorate)
            handler_dir = os.path.dirname(handler_file)
            if not os.path.isdir(handler_dir):
                os.makedirs(handler_dir)
            with codecs.open(handler_file, "w", "utf-8") as f:
                f.write(handler)
            self._notify_create(target, handler, handler_file)

        entry = (target, handler, handler_file)
        self._handler_by_id[target.identifier] = entry
        self._handler_by_file[handler_file] = entry

        self._ensure_monitor(handler_file)

        return handler

    def update_handler(self, target: TraceTarget, handler: str) -> None:
        _, _, handler_file = self._handler_by_id.get(target.identifier)
        entry = (target, handler, handler_file)
        self._handler_by_id[target.identifier] = entry
        self._handler_by_file[handler_file] = entry
        self._notify_update(target, handler, handler_file)

        Path(handler_file).write_text(handler, encoding="utf-8")

    def _ensure_monitor(self, handler_file) -> None:
        handler_dir = os.path.dirname(handler_file)
        monitor = self._repo_monitors.get(handler_dir)
        if monitor is None:
            monitor = frida.FileMonitor(handler_dir)
            monitor.on("change", self._on_change)
            self._repo_monitors[handler_dir] = monitor

    def commit_handlers(self) -> None:
        for monitor in self._repo_monitors.values():
            monitor.enable()

    def _on_change(self, changed_file, other_file, event_type) -> None:
        if changed_file not in self._handler_by_file or event_type == "changes-done-hint":
            return
        self._changed_files.add(changed_file)
        self._last_change_id += 1
        change_id = self._last_change_id
        self._reactor.schedule(lambda: self._sync_handlers(change_id), delay=0.05)

    def _sync_handlers(self, change_id) -> None:
        if change_id != self._last_change_id:
            return
        changes = self._changed_files.copy()
        self._changed_files.clear()
        for changed_handler_file in changes:
            (target, old_handler, handler_file) = self._handler_by_file[changed_handler_file]
            with codecs.open(handler_file, "r", "utf-8") as f:
                new_handler = f.read()
            changed = new_handler != old_handler
            if changed:
                entry = (target, new_handler, handler_file)
                self._handler_by_id[target.identifier] = entry
                self._handler_by_file[handler_file] = entry
                self._notify_update(target, new_handler, handler_file)


class InitScript:
    def __init__(self, filename, source) -> None:
        self.filename = filename
        self.source = source


class OutputFile:
    def __init__(self, filename: str) -> None:
        self._fd = codecs.open(filename, "wb", "utf-8")

    def close(self) -> None:
        self._fd.close()

    def append(self, message: str) -> None:
        self._fd.write(message)
        self._fd.flush()


class UI:
    def on_script_created(self, script: frida.core.Script) -> None:
        pass

    def on_trace_progress(self, status) -> None:
        pass

    def on_trace_warning(self, message):
        pass

    def on_trace_error(self, message) -> None:
        pass

    def on_trace_events(self, events) -> None:
        pass

    def on_trace_handler_create(self, target, handler, source) -> None:
        pass

    def on_trace_handler_load(self, target, handler, source) -> None:
        pass


def to_filename(name: str) -> str:
    result = ""
    for c in name:
        if c.isalnum() or c == ".":
            result += c
        else:
            result += "_"
    return result


def to_handler_filename(name: str) -> str:
    full_filename = to_filename(name)
    if len(full_filename) <= 41:
        return full_filename + ".js"
    crc = binascii.crc32(full_filename.encode())
    return full_filename[0:32] + "_%08x.js" % crc


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass
